# Copyright (c) 2014 X-IO Technologies.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_log import log as logging

from cinder import context
from cinder import exception
from cinder import test
from cinder.tests.unit import utils
from cinder.volume.drivers import xio
from cinder.volume import qos_specs
from cinder.volume import volume_types

LOG = logging.getLogger("cinder.volume.driver")

ISE_IP1 = '10.12.12.1'
ISE_IP2 = '10.11.12.2'
ISE_ISCSI_IP1 = '1.2.3.4'
ISE_ISCSI_IP2 = '1.2.3.5'

ISE_GID = 'isegid'
ISE_IQN = ISE_GID
ISE_WWN1 = ISE_GID + '1'
ISE_WWN2 = ISE_GID + '2'
ISE_WWN3 = ISE_GID + '3'
ISE_WWN4 = ISE_GID + '4'
ISE_TARGETS = [ISE_WWN1, ISE_WWN2, ISE_WWN3, ISE_WWN4]
ISE_INIT_TARGET_MAP = {'init_wwn1': ISE_TARGETS,
                       'init_wwn2': ISE_TARGETS}

VOLUME_SIZE = 10
NEW_VOLUME_SIZE = 20

VOLUME1 = {'id': '1', 'name': 'volume1',
           'size': VOLUME_SIZE, 'volume_type_id': 'type1'}

VOLUME2 = {'id': '2', 'name': 'volume2',
           'size': VOLUME_SIZE, 'volume_type_id': 'type2',
           'provider_auth': 'CHAP abc abc'}

VOLUME3 = {'id': '3', 'name': 'volume3',
           'size': VOLUME_SIZE, 'volume_type_id': None}

SNAPSHOT1 = {'name': 'snapshot1',
             'volume_name': VOLUME1['name'],
             'volume_type_id': 'type3'}

CLONE1 = {'id': '3', 'name': 'clone1',
          'size': VOLUME_SIZE, 'volume_type_id': 'type4'}

HOST1 = 'host1'

HOST2 = 'host2'

ISCSI_CONN1 = {'initiator': 'init_iqn1',
               'host': HOST1}

ISCSI_CONN2 = {'initiator': 'init_iqn2',
               'host': HOST2}

FC_CONN1 = {'wwpns': ['init_wwn1', 'init_wwn2'],
            'host': HOST1}

FC_CONN2 = {'wwpns': ['init_wwn3', 'init_wwn4'],
            'host': HOST2}

ISE_HTTP_IP = 'http://' + ISE_IP1

ISE_HOST_LOCATION = '/storage/hosts/1'
ISE_HOST_LOCATION_URL = ISE_HTTP_IP + ISE_HOST_LOCATION

ISE_VOLUME1_LOCATION = '/storage/volumes/volume1'
ISE_VOLUME1_LOCATION_URL = ISE_HTTP_IP + ISE_VOLUME1_LOCATION
ISE_VOLUME2_LOCATION = '/storage/volumes/volume2'
ISE_VOLUME2_LOCATION_URL = ISE_HTTP_IP + ISE_VOLUME2_LOCATION
ISE_VOLUME3_LOCATION = '/storage/volumes/volume3'
ISE_VOLUME3_LOCATION_URL = ISE_HTTP_IP + ISE_VOLUME3_LOCATION

ISE_SNAPSHOT_LOCATION = '/storage/volumes/snapshot1'
ISE_SNAPSHOT_LOCATION_URL = ISE_HTTP_IP + ISE_SNAPSHOT_LOCATION

ISE_CLONE_LOCATION = '/storage/volumes/clone1'
ISE_CLONE_LOCATION_URL = ISE_HTTP_IP + ISE_CLONE_LOCATION

ISE_ALLOCATION_LOCATION = '/storage/allocations/a1'
ISE_ALLOCATION_LOCATION_URL = ISE_HTTP_IP + ISE_ALLOCATION_LOCATION

ISE_GET_QUERY_XML =\
    """<array>
        <globalid>ABC12345</globalid>
        <capabilities>
            <capability value="3" string="Storage" type="source"/>
            <capability value="49003" string="Volume Affinity"/>
            <capability value="49004" string="Volume Quality of Service IOPS"/>
            <capability value="49005" string="Thin Provisioning"/>
            <capability value="49006" string="Clones" type="source"/>
        </capabilities>
        <controllers>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="1"/>
            </controller>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="0"/>
            </controller>
        </controllers>
       </array>""" % (ISE_IP1, ISE_IP2)

ISE_GET_QUERY_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_XML.split())}

ISE_GET_QUERY_NO_CAP_XML =\
    """<array>
        <globalid>ABC12345</globalid>
        <controllers>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="1"/>
            </controller>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="0"/>
            </controller>
        </controllers>
       </array>""" % (ISE_IP1, ISE_IP2)

ISE_GET_QUERY_NO_CAP_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_NO_CAP_XML.split())}

ISE_GET_QUERY_NO_CTRL_XML =\
    """<array>
        <globalid>ABC12345</globalid>
        <capabilities>
            <capability value="3" string="Storage" type="source"/>
            <capability value="49003" string="Volume Affinity"/>
            <capability value="49004" string="Volume Quality of Service IOPS"/>
            <capability value="49005" string="Thin Provisioning"/>
            <capability value="49006" string="Clones" type="source"/>
            <capability value="49007" string="Thin clones" type="source"/>
            <capability value="49007" string="Thin clones" type="source"/>
        </capabilities>
       </array>"""

ISE_GET_QUERY_NO_CTRL_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_NO_CTRL_XML.split())}

ISE_GET_QUERY_NO_IP_XML =\
    """<array>
        <globalid>ABC12345</globalid>
        <capabilities>
            <test value="1"/>
            <capability value="3" string="Storage" type="source"/>
            <capability value="49003" string="Volume Affinity"/>
            <capability value="49004" string="Volume Quality of Service IOPS"/>
            <capability value="49005" string="Thin Provisioning"/>
            <capability value="49006" string="Clones" type="source"/>
            <capability value="49007" string="Thin clones" type="source"/>
            <capability value="49007" string="Thin clones" type="source"/>
        </capabilities>
        <controllers>
            <test value="2"/>
            <controller>
                <rank value="1"/>
            </controller>
            <controller>
                <rank value="0"/>
            </controller>
        </controllers>
       </array>"""

ISE_GET_QUERY_NO_IP_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_NO_IP_XML.split())}

ISE_GET_QUERY_NO_GID_XML =\
    """<array>
        <capabilities>
            <capability value="3" string="Storage" type="source"/>
            <capability value="49003" string="Volume Affinity"/>
            <capability value="49004" string="Volume Quality of Service IOPS"/>
            <capability value="49005" string="Thin Provisioning"/>
            <capability value="49006" string="Clones" type="source"/>
        </capabilities>
        <controllers>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="1"/>
            </controller>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="0"/>
            </controller>
        </controllers>
       </array>""" % (ISE_IP1, ISE_IP2)

ISE_GET_QUERY_NO_GID_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_NO_GID_XML.split())}

ISE_GET_QUERY_NO_CLONE_XML =\
    """<array>
        <globalid>ABC12345</globalid>
        <capabilities>
            <capability value="3" string="Storage" type="source"/>
            <capability value="49003" string="Volume Affinity"/>
            <capability value="49004" string="Volume Quality of Service IOPS"/>
            <capability value="49005" string="Thin Provisioning"/>
        </capabilities>
        <controllers>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="1"/>
            </controller>
            <controller>
                <ipaddress>%s</ipaddress>
                <rank value="0"/>
            </controller>
        </controllers>
       </array>""" % (ISE_IP1, ISE_IP2)

ISE_GET_QUERY_NO_CLONE_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_QUERY_NO_CLONE_XML.split())}

ISE_GET_STORAGE_POOLS_XML =\
    """
    <pools>
        <pool>
            <name>Pool 1</name>
            <id>1</id>
            <status value="0" string="Operational">
                <details value="0x00000000">
                    <detail>None</detail>
                </details>
            </status>
            <available total="60">
                <byredundancy>
                    <raid-0>60</raid-0>
                    <raid-1>30</raid-1>
                    <raid-5>45</raid-5>
                </byredundancy>
            </available>
            <used total="40">
                <byredundancy>
                    <raid-0>0</raid-0>
                    <raid-1>40</raid-1>
                    <raid-5>0</raid-5>
                </byredundancy>
            </used>
            <media>
                <medium>
                    <health>100</health>
                    <tier value="4" string="Hybrid"/>
                </medium>
            </media>
            <volumes>
                <volume>
                    <globalid>volgid</globalid>
                </volume>
                <volume>
                    <globalid>volgid2</globalid>
                </volume>
            </volumes>
        </pool>
    </pools>
    """

ISE_GET_STORAGE_POOLS_RESP =\
    {'status': 200,
     'location': 'Pool location',
     'content': " ".join(ISE_GET_STORAGE_POOLS_XML.split())}

ISE_GET_VOL_STATUS_NO_VOL_NODE_XML =\
    """<volumes></volumes>"""

ISE_GET_VOL_STATUS_NO_VOL_NODE_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_VOLUME1_LOCATION_URL,
     'content': " ".join(ISE_GET_VOL_STATUS_NO_VOL_NODE_XML.split())}

ISE_GET_VOL_STATUS_NO_STATUS_XML =\
    """<volumes>
        <volume self="%s">
        </volume>
    </volumes>""" % (ISE_VOLUME1_LOCATION_URL)

ISE_GET_VOL_STATUS_NO_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_VOLUME1_LOCATION_URL,
     'content': " ".join(ISE_GET_VOL_STATUS_NO_STATUS_XML.split())}

ISE_GET_VOL1_STATUS_XML =\
    """<volumes>
        <volume self="%s">
            <status value="0" string="Operational">
                <details>
                    <detail>Prepared</detail>
                </details>
            </status>
            <size>10</size>
        </volume>
    </volumes>""" % (ISE_VOLUME1_LOCATION_URL)

ISE_GET_VOL1_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_VOLUME1_LOCATION_URL,
     'content': " ".join(ISE_GET_VOL1_STATUS_XML.split())}

ISE_GET_VOL2_STATUS_XML =\
    """<volumes>
        <volume self="%s">
            <status value="0" string="Operational">
                <details>
                    <detail>Prepared</detail>
                </details>
            </status>
        </volume>
    </volumes>""" % (ISE_VOLUME2_LOCATION_URL)

ISE_GET_VOL2_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_VOLUME2_LOCATION_URL,
     'content': " ".join(ISE_GET_VOL2_STATUS_XML.split())}

ISE_GET_VOL3_STATUS_XML =\
    """<volumes>
        <volume self="%s">
            <status value="0" string="Operational">
                <details>
                    <detail>Prepared</detail>
                </details>
            </status>
        </volume>
    </volumes>""" % (ISE_VOLUME3_LOCATION_URL)

ISE_GET_VOL3_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_VOLUME3_LOCATION_URL,
     'content': " ".join(ISE_GET_VOL3_STATUS_XML.split())}

ISE_GET_SNAP1_STATUS_XML =\
    """<volumes>
        <volume self="%s">
            <status value="0" string="Operational">
                <details>
                    <detail>Prepared</detail>
                </details>
            </status>
        </volume>
    </volumes>""" % (ISE_SNAPSHOT_LOCATION_URL)

ISE_GET_SNAP1_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_SNAPSHOT_LOCATION_URL,
     'content': " ".join(ISE_GET_SNAP1_STATUS_XML.split())}

ISE_GET_CLONE1_STATUS_XML =\
    """<volumes>
        <volume self="%s">
            <status value="0" string="Operational">
                <details>
                    <detail>Prepared</detail>
                </details>
            </status>
        </volume>
    </volumes>""" % (ISE_CLONE_LOCATION_URL)

ISE_GET_CLONE1_STATUS_RESP =\
    {'status': 200,
     'location': 'u%s' % ISE_CLONE_LOCATION_URL,
     'content': " ".join(ISE_GET_CLONE1_STATUS_XML.split())}

ISE_CREATE_VOLUME_XML = """<volume/>"""

ISE_CREATE_VOLUME_RESP =\
    {'status': 201,
     'location': ISE_VOLUME1_LOCATION_URL,
     'content': " ".join(ISE_CREATE_VOLUME_XML.split())}

ISE_GET_IONETWORKS_XML =\
    """<chap>
        <chapin value="0" string="disabled">
            <username/>
            <password/>
        </chapin>
        <chapout value="0" string="disabled">
            <username/>
            <password/>
        </chapout>
       </chap>"""

ISE_GET_IONETWORKS_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_IONETWORKS_XML.split())}

ISE_GET_IONETWORKS_CHAP_XML =\
    """<chap>
        <chapin value="1" string="disabled">
            <username>abc</username>
            <password>abc</password>
        </chapin>
        <chapout value="0" string="disabled">
            <username/>
            <password/>
        </chapout>
       </chap>"""

ISE_GET_IONETWORKS_CHAP_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_IONETWORKS_CHAP_XML.split())}

ISE_DELETE_VOLUME_XML = """<volumes/>"""

ISE_DELETE_VOLUME_RESP =\
    {'status': 204,
     'location': '',
     'content': " ".join(ISE_DELETE_VOLUME_XML.split())}

ISE_GET_ALLOC_WITH_EP_XML =\
    """<allocations>
        <allocation self="%s">
            <volume>
                <volumename>%s</volumename>
            </volume>
            <endpoints>
                <hostname>%s</hostname>
            </endpoints>
            <lun>1</lun>
        </allocation>
       </allocations>""" %\
    (ISE_ALLOCATION_LOCATION_URL, VOLUME1['name'], HOST1)

ISE_GET_ALLOC_WITH_EP_RESP =\
    {'status': 200,
     'location': ISE_ALLOCATION_LOCATION_URL,
     'content': " ".join(ISE_GET_ALLOC_WITH_EP_XML.split())}

ISE_GET_ALLOC_WITH_NO_ALLOC_XML =\
    """<allocations self="%s"/>""" % ISE_ALLOCATION_LOCATION_URL

ISE_GET_ALLOC_WITH_NO_ALLOC_RESP =\
    {'status': 200,
     'location': ISE_ALLOCATION_LOCATION_URL,
     'content': " ".join(ISE_GET_ALLOC_WITH_NO_ALLOC_XML.split())}

ISE_DELETE_ALLOC_XML = """<allocations/>"""

ISE_DELETE_ALLOC_RESP =\
    {'status': 204,
     'location': '',
     'content': " ".join(ISE_DELETE_ALLOC_XML.split())}

ISE_GET_HOSTS_NOHOST_XML =\
    """<hosts self="http://ip/storage/hosts"/>"""

ISE_GET_HOSTS_NOHOST_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_HOSTS_NOHOST_XML.split())}

ISE_GET_HOSTS_HOST1_XML =\
    """<hosts self="http://ip/storage/hosts">
        <host self="http://ip/storage/hosts/1">
            <type>"OPENSTACK"</type>
            <name>%s</name>
            <id>1</id>
            <endpoints self="http://ip/storage/endpoints">
                <endpoint self="http://ip/storage/endpoints/ep1">
                    <globalid>init_wwn1</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep2">
                    <globalid>init_wwn2</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep1">
                    <globalid>init_iqn1</globalid>
                </endpoint>
            </endpoints>
        </host>
       </hosts>""" % HOST1

ISE_GET_HOSTS_HOST1_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_HOSTS_HOST1_XML.split())}

ISE_GET_HOSTS_HOST1_HOST_TYPE_XML =\
    """<hosts self="http://ip/storage/hosts">
        <host self="http://ip/storage/hosts/1">
            <type>"WINDOWS"</type>
            <name>%s</name>
            <id>1</id>
            <endpoints self="http://ip/storage/endpoints">
                <endpoint self="http://ip/storage/endpoints/ep1">
                    <globalid>init_wwn1</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep2">
                    <globalid>init_wwn2</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep1">
                    <globalid>init_iqn1</globalid>
                </endpoint>
            </endpoints>
        </host>
       </hosts>""" % HOST1

ISE_GET_HOSTS_HOST1_HOST_TYPE_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_HOSTS_HOST1_HOST_TYPE_XML.split())}

ISE_GET_HOSTS_HOST2_XML =\
    """<hosts self="http://ip/storage/hosts">
        <host self="http://ip/storage/hosts/2">
            <name>%s</name>
            <id>2</id>
            <endpoints self="http://ip/storage/endpoints">
                <endpoint self="http://ip/storage/endpoints/ep3">
                    <globalid>init_wwn3</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep4">
                    <globalid>init_wwn4</globalid>
                </endpoint>
                <endpoint self="http://ip/storage/endpoints/ep3">
                    <globalid>init_iqn2</globalid>
                </endpoint>
            </endpoints>
        </host>
       </hosts>""" % HOST2

ISE_GET_HOSTS_HOST2_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_HOSTS_HOST2_XML.split())}

ISE_CREATE_HOST_XML =\
    """<hosts self="http://ip/storage/hosts"/>"""

ISE_CREATE_HOST_RESP =\
    {'status': 201,
     'location': 'http://ip/storage/hosts/host1',
     'content': " ".join(ISE_CREATE_HOST_XML.split())}

ISE_CREATE_ALLOC_XML =\
    """<allocations self="http://ip/storage/allocations"/>"""

ISE_CREATE_ALLOC_RESP =\
    {'status': 201,
     'location': ISE_ALLOCATION_LOCATION_URL,
     'content': " ".join(ISE_CREATE_ALLOC_XML.split())}

ISE_GET_ENDPOINTS_XML =\
    """<endpoints self="http://ip/storage/endpoints">
        <endpoint type="array" self="http://ip/storage/endpoints/isegid">
                <globalid>isegid</globalid>
                <protocol>iSCSI</protocol>
                <array self="http://ip/storage/arrays/ise1">
                    <globalid>ise1</globalid>
                </array>
                <host/>
                <allocations self="http://ip/storage/allocations">
                    <allocation self="%s">
                        <globalid>
                            a1
                        </globalid>
                    </allocation>
                </allocations>
            </endpoint>
        <endpoint type="array" self="http://ip/storage/endpoints/isegid">
                <globalid>isegid</globalid>
                <protocol>Fibre Channel</protocol>
                <array self="http://ip/storage/arrays/ise1">
                    <globalid>ise1</globalid>
                </array>
                <host/>
                <allocations self="http://ip/storage/allocations">
                    <allocation self="%s">
                        <globalid>
                            a1
                        </globalid>
                    </allocation>
                </allocations>
            </endpoint>
        </endpoints>""" % (ISE_ALLOCATION_LOCATION_URL,
                           ISE_ALLOCATION_LOCATION_URL)

ISE_GET_ENDPOINTS_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_ENDPOINTS_XML.split())}

ISE_GET_CONTROLLERS_XML =\
    """<controllers self="http://ip/storage/arrays/controllers">
        <controller>
            <status/>
            <ioports>
                <ioport>
                    <ipaddresses>
                        <ipaddress>%s</ipaddress>
                    </ipaddresses>
                    <endpoint>
                        <globalid>isegid</globalid>
                    </endpoint>
                </ioport>
            </ioports>
            <fcports>
                <fcport>
                    <wwn>%s</wwn>
                </fcport>
                <fcport>
                    <wwn>%s</wwn>
                </fcport>
            </fcports>
        </controller>
        <controller>
            <status/>
            <ioports>
                <ioport>
                    <ipaddresses>
                        <ipaddress>%s</ipaddress>
                    </ipaddresses>
                    <endpoint>
                        <globalid>isegid</globalid>
                    </endpoint>
                </ioport>
            </ioports>
            <fcports>
                <fcport>
                    <wwn>%s</wwn>
                </fcport>
                <fcport>
                    <wwn>%s</wwn>
                </fcport>
            </fcports>
        </controller>
       </controllers>""" % (ISE_ISCSI_IP1, ISE_WWN1, ISE_WWN2,
                            ISE_ISCSI_IP2, ISE_WWN3, ISE_WWN4)

ISE_GET_CONTROLLERS_RESP =\
    {'status': 200,
     'location': '',
     'content': " ".join(ISE_GET_CONTROLLERS_XML.split())}

ISE_CREATE_SNAPSHOT_XML = """<snapshot/>"""

ISE_CREATE_SNAPSHOT_RESP =\
    {'status': 201,
     'location': ISE_SNAPSHOT_LOCATION_URL,
     'content': " ".join(ISE_CREATE_SNAPSHOT_XML.split())}

ISE_PREP_SNAPSHOT_XML = """<snapshot/>"""

ISE_PREP_SNAPSHOT_RESP =\
    {'status': 202,
     'location': ISE_SNAPSHOT_LOCATION_URL,
     'content': " ".join(ISE_PREP_SNAPSHOT_XML.split())}

ISE_MODIFY_VOLUME_XML = """<volume/>"""

ISE_MODIFY_VOLUME_RESP =\
    {'status': 201,
     'location': ISE_VOLUME1_LOCATION_URL,
     'content': " ".join(ISE_MODIFY_VOLUME_XML.split())}

ISE_MODIFY_HOST_XML = """<host/>"""

ISE_MODIFY_HOST_RESP =\
    {'status': 201,
     'location': ISE_HOST_LOCATION_URL,
     'content': " ".join(ISE_MODIFY_HOST_XML.split())}

ISE_BAD_CONNECTION_RESP =\
    {'status': 0,
     'location': '',
     'content': " "}

ISE_400_RESP =\
    {'status': 400,
     'location': '',
     'content': ""}

ISE_GET_VOL_STATUS_404_XML = \
    """<response value="404" index="3">VOLUME not found.</response>"""

ISE_GET_VOL_STATUS_404_RESP =\
    {'status': 404,
     'location': '',
     'content': " ".join(ISE_GET_VOL_STATUS_404_XML.split())}

ISE_400_INVALID_STATE_XML = \
    """<response value="400">Not in a valid state.</response>"""

ISE_400_INVALID_STATE_RESP =\
    {'status': 400,
     'location': '',
     'content': " ".join(ISE_400_INVALID_STATE_XML.split())}

ISE_409_CONFLICT_XML = \
    """<response value="409">Conflict</response>"""

ISE_409_CONFLICT_RESP =\
    {'status': 409,
     'location': '',
     'content': " ".join(ISE_409_CONFLICT_XML.split())}


DRIVER = "cinder.volume.drivers.xio.XIOISEDriver"


@mock.patch(DRIVER + "._opener", autospec=True)
class XIOISEDriverTestCase(object):

    # Test cases for X-IO volume driver

    def setUp(self):
        super(XIOISEDriverTestCase, self).setUp()

        # set good default values
        self.configuration = mock.Mock()
        self.configuration.san_ip = ISE_IP1
        self.configuration.san_user = 'fakeuser'
        self.configuration.san_password = 'fakepass'
        self.configuration.iscsi_ip_address = ISE_ISCSI_IP1
        self.configuration.driver_use_ssl = False
        self.configuration.ise_completion_retries = 30
        self.configuration.ise_connection_retries = 5
        self.configuration.ise_retry_interval = 1
        self.configuration.volume_backend_name = 'ise1'
        self.driver = None
        self.protocol = ''
        self.connector = None
        self.connection_failures = 0
        self.hostgid = ''
        self.use_response_table = 1

    def setup_test(self, protocol):
        self.protocol = protocol

        # set good default values
        if self.protocol == 'iscsi':
            self.configuration.ise_protocol = protocol
            self.connector = ISCSI_CONN1
            self.hostgid = self.connector['initiator']
        elif self.protocol == 'fibre_channel':
            self.configuration.ise_protocol = protocol
            self.connector = FC_CONN1
            self.hostgid = self.connector['wwpns'][0]

    def setup_driver(self):
        # this setups up driver object with previously set configuration values
        if self.configuration.ise_protocol == 'iscsi':
            self.driver =\
                xio.XIOISEISCSIDriver(configuration=self.configuration)
        elif self.configuration.ise_protocol == 'fibre_channel':
            self.driver =\
                xio.XIOISEFCDriver(configuration=self.configuration)
        elif self.configuration.ise_protocol == 'test_prot':
            # if test_prot specified override with correct protocol
            # used to bypass protocol specific driver
            self.configuration.ise_protocol = self.protocol
            self.driver = xio.XIOISEDriver(configuration=self.configuration)
        else:
            # Invalid protocol type
            raise exception.Invalid()

#################################
#         UNIT TESTS            #
#################################
    def test_do_setup(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP])
        self.driver.do_setup(None)

    def test_negative_do_setup_no_clone_support(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_NO_CLONE_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.do_setup, None)

    def test_negative_do_setup_no_capabilities(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_NO_CAP_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.do_setup, None)

    def test_negative_do_setup_no_ctrl(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_NO_CTRL_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.do_setup, None)

    def test_negative_do_setup_no_ipaddress(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_NO_IP_RESP])
        self.driver.do_setup(None)

    def test_negative_do_setup_bad_globalid_none(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_NO_GID_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.do_setup, None)

    def test_check_for_setup_error(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP])
        self.setup_driver()
        self.driver.check_for_setup_error()

    def test_negative_do_setup_bad_ip(self, mock_req):
        # set san_ip to bad value
        self.configuration.san_ip = ''
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP])
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.check_for_setup_error)

    def test_negative_do_setup_bad_user_blank(self, mock_req):
        # set san_user to bad value
        self.configuration.san_login = ''
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP])
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.check_for_setup_error)

    def test_negative_do_setup_bad_password_blank(self, mock_req):
        # set san_password to bad value
        self.configuration.san_password = ''
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP])
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.check_for_setup_error)

    def test_get_volume_stats(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_STORAGE_POOLS_RESP])

        backend_name = self.configuration.volume_backend_name
        if self.configuration.ise_protocol == 'iscsi':
            protocol = 'iSCSI'
        else:
            protocol = 'fibre_channel'
        exp_result = {'vendor_name': "X-IO",
                      'driver_version': "1.1.4",
                      'volume_backend_name': backend_name,
                      'reserved_percentage': 0,
                      'total_capacity_gb': 100,
                      'free_capacity_gb': 60,
                      'QoS_support': True,
                      'affinity': True,
                      'thin': False,
                      'pools': [{'pool_ise_name': "Pool 1",
                                 'pool_name': "1",
                                 'status': "Operational",
                                 'status_details': "None",
                                 'free_capacity_gb': 60,
                                 'free_capacity_gb_raid_0': 60,
                                 'free_capacity_gb_raid_1': 30,
                                 'free_capacity_gb_raid_5': 45,
                                 'allocated_capacity_gb': 40,
                                 'allocated_capacity_gb_raid_0': 0,
                                 'allocated_capacity_gb_raid_1': 40,
                                 'allocated_capacity_gb_raid_5': 0,
                                 'health': 100,
                                 'media': "Hybrid",
                                 'total_capacity_gb': 100,
                                 'QoS_support': True,
                                 'reserved_percentage': 0}],
                      'active_volumes': 2,
                      'storage_protocol': protocol}

        act_result = self.driver.get_volume_stats(True)
        self.assertDictMatch(exp_result, act_result)

    def test_get_volume_stats_ssl(self, mock_req):
        self.configuration.driver_use_ssl = True
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_STORAGE_POOLS_RESP])
        self.driver.get_volume_stats(True)

    def test_negative_get_volume_stats_bad_primary(self, mock_req):
        self.configuration.ise_connection_retries = 1
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_BAD_CONNECTION_RESP,
                                     ISE_GET_STORAGE_POOLS_RESP])
        self.driver.get_volume_stats(True)

    def test_create_volume(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        VOLUME1['volume_type_id'] = type_ref['id']
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_GET_IONETWORKS_RESP])
            exp_result = {}
            exp_result = {"provider_auth": ""}
            act_result = self.driver.create_volume(VOLUME1)
            self.assertDictMatch(exp_result, act_result)
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP])
            self.driver.create_volume(VOLUME1)

    def test_create_volume_chap(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        VOLUME1['volume_type_id'] = type_ref['id']
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_GET_IONETWORKS_CHAP_RESP])
            exp_result = {}
            exp_result = {"provider_auth": "CHAP abc abc"}
            act_result = self.driver.create_volume(VOLUME1)
            self.assertDictMatch(exp_result, act_result)
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP])
            self.driver.create_volume(VOLUME1)

    def test_create_volume_type_none(self, mock_req):
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_GET_IONETWORKS_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_CREATE_VOLUME_RESP,
                                         ISE_GET_VOL1_STATUS_RESP])
        self.driver.create_volume(VOLUME3)

    def test_delete_volume(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_DELETE_ALLOC_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_DELETE_VOLUME_RESP,
                                     ISE_GET_VOL_STATUS_404_RESP])
        self.setup_driver()
        self.driver.delete_volume(VOLUME1)

    def test_delete_volume_delayed(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_DELETE_ALLOC_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_DELETE_VOLUME_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_GET_VOL_STATUS_404_RESP])
        self.setup_driver()
        self.driver.delete_volume(VOLUME1)

    def test_delete_volume_timeout(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_DELETE_ALLOC_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_DELETE_VOLUME_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_GET_VOL1_STATUS_RESP])

        self.configuration.ise_completion_retries = 3
        self.setup_driver()
        self.driver.delete_volume(VOLUME1)

    def test_delete_volume_none_existing(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_DELETE_ALLOC_RESP,
                                     ISE_GET_VOL1_STATUS_RESP])
        self.setup_driver()
        self.driver.delete_volume(VOLUME2)

    def test_initialize_connection_positive(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_HOSTS_HOST2_RESP,
                                     ISE_CREATE_HOST_RESP,
                                     ISE_GET_HOSTS_HOST1_RESP,
                                     ISE_CREATE_ALLOC_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_GET_CONTROLLERS_RESP])
        self.setup_driver()

        exp_result = {}
        if self.configuration.ise_protocol == 'iscsi':
            exp_result = {"driver_volume_type": "iscsi",
                          "data": {"target_lun": 1,
                                   "volume_id": '1',
                                   "target_discovered": False,
                                   "target_iqn": ISE_IQN,
                                   "target_portal": ISE_ISCSI_IP1 + ":3260"}}
        elif self.configuration.ise_protocol == 'fibre_channel':
            exp_result = {"driver_volume_type": "fibre_channel",
                          "data": {"target_lun": 1,
                                   "volume_id": '1',
                                   "target_discovered": True,
                                   "initiator_target_map": ISE_INIT_TARGET_MAP,
                                   "target_wwn": ISE_TARGETS}}

        act_result =\
            self.driver.initialize_connection(VOLUME1, self.connector)
        self.assertDictMatch(exp_result, act_result)

    def test_initialize_connection_positive_host_type(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_HOSTS_HOST1_HOST_TYPE_RESP,
                                     ISE_MODIFY_HOST_RESP,
                                     ISE_CREATE_ALLOC_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_GET_CONTROLLERS_RESP])
        self.setup_driver()

        exp_result = {}
        if self.configuration.ise_protocol == 'iscsi':
            exp_result = {"driver_volume_type": "iscsi",
                          "data": {"target_lun": 1,
                                   "volume_id": '1',
                                   "target_discovered": False,
                                   "target_iqn": ISE_IQN,
                                   "target_portal": ISE_ISCSI_IP1 + ":3260"}}
        elif self.configuration.ise_protocol == 'fibre_channel':
            exp_result = {"driver_volume_type": "fibre_channel",
                          "data": {"target_lun": 1,
                                   "volume_id": '1',
                                   "target_discovered": True,
                                   "initiator_target_map": ISE_INIT_TARGET_MAP,
                                   "target_wwn": ISE_TARGETS}}

        act_result =\
            self.driver.initialize_connection(VOLUME1, self.connector)
        self.assertDictMatch(exp_result, act_result)

    def test_initialize_connection_positive_chap(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_HOSTS_HOST2_RESP,
                                     ISE_CREATE_HOST_RESP,
                                     ISE_GET_HOSTS_HOST1_RESP,
                                     ISE_CREATE_ALLOC_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_GET_CONTROLLERS_RESP])
        self.setup_driver()
        exp_result = {}
        if self.configuration.ise_protocol == 'iscsi':
            exp_result = {"driver_volume_type": "iscsi",
                          "data": {"target_lun": 1,
                                   "volume_id": '2',
                                   "target_discovered": False,
                                   "target_iqn": ISE_IQN,
                                   "target_portal": ISE_ISCSI_IP1 + ":3260",
                                   'auth_method': 'CHAP',
                                   'auth_username': 'abc',
                                   'auth_password': 'abc'}}
        elif self.configuration.ise_protocol == 'fibre_channel':
            exp_result = {"driver_volume_type": "fibre_channel",
                          "data": {"target_lun": 1,
                                   "volume_id": '2',
                                   "target_discovered": True,
                                   "initiator_target_map": ISE_INIT_TARGET_MAP,
                                   "target_wwn": ISE_TARGETS}}

        act_result =\
            self.driver.initialize_connection(VOLUME2, self.connector)
        self.assertDictMatch(exp_result, act_result)

    def test_initialize_connection_negative_no_host(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_HOSTS_HOST2_RESP,
                                     ISE_CREATE_HOST_RESP,
                                     ISE_GET_HOSTS_HOST2_RESP])
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.initialize_connection,
                          VOLUME2, self.connector)

    def test_initialize_connection_negative_host_type(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_HOSTS_HOST1_HOST_TYPE_RESP,
                                     ISE_400_RESP])
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.initialize_connection,
                          VOLUME2, self.connector)

    def test_terminate_connection_positive(self, mock_req):
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_GET_ALLOC_WITH_EP_RESP,
                                         ISE_DELETE_ALLOC_RESP,
                                         ISE_GET_ALLOC_WITH_EP_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_DELETE_ALLOC_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_GET_ALLOC_WITH_EP_RESP,
                                         ISE_DELETE_ALLOC_RESP,
                                         ISE_GET_ALLOC_WITH_EP_RESP,
                                         ISE_GET_CONTROLLERS_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_DELETE_ALLOC_RESP])
        self.driver.terminate_connection(VOLUME1, self.connector)

    def test_terminate_connection_positive_noalloc(self, mock_req):
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_GET_ALLOC_WITH_NO_ALLOC_RESP,
                                         ISE_GET_ALLOC_WITH_NO_ALLOC_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_DELETE_ALLOC_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_GET_ALLOC_WITH_NO_ALLOC_RESP,
                                         ISE_GET_ALLOC_WITH_NO_ALLOC_RESP,
                                         ISE_GET_CONTROLLERS_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_DELETE_ALLOC_RESP])
        self.driver.terminate_connection(VOLUME1, self.connector)

    def test_negative_terminate_connection_bad_host(self, mock_req):
        self.setup_driver()
        test_connector = {}
        if self.configuration.ise_protocol == 'iscsi':
            test_connector['initiator'] = 'bad_iqn'
            test_connector['host'] = ''
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            test_connector['wwpns'] = 'bad_wwn'
            test_connector['host'] = ''
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_HOSTS_HOST1_RESP,
                                         ISE_GET_CONTROLLERS_RESP])

        self.driver.terminate_connection(VOLUME1, test_connector)

    def test_create_snapshot(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        SNAPSHOT1['volume_type_id'] = type_ref['id']

        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_PREP_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP,
                                     ISE_CREATE_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP])
        self.setup_driver()
        self.driver.create_snapshot(SNAPSHOT1)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=utils.ZeroIntervalLoopingCall)
    def test_negative_create_snapshot_invalid_state_recover(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        SNAPSHOT1['volume_type_id'] = type_ref['id']

        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_400_INVALID_STATE_RESP,
                                     ISE_PREP_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP,
                                     ISE_CREATE_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP])
        self.setup_driver()
        self.driver.create_snapshot(SNAPSHOT1)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall',
                new=utils.ZeroIntervalLoopingCall)
    def test_negative_create_snapshot_invalid_state_norecover(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        SNAPSHOT1['volume_type_id'] = type_ref['id']

        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_400_INVALID_STATE_RESP,
                                     ISE_400_INVALID_STATE_RESP,
                                     ISE_400_INVALID_STATE_RESP,
                                     ISE_400_INVALID_STATE_RESP,
                                     ISE_400_INVALID_STATE_RESP])
        self.configuration.ise_completion_retries = 5
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.create_snapshot, SNAPSHOT1)

    def test_negative_create_snapshot_conflict(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        SNAPSHOT1['volume_type_id'] = type_ref['id']

        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_409_CONFLICT_RESP])
        self.configuration.ise_completion_retries = 1
        self.setup_driver()
        self.assertRaises(exception.XIODriverException,
                          self.driver.create_snapshot, SNAPSHOT1)

    def test_delete_snapshot(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_ALLOC_WITH_EP_RESP,
                                     ISE_DELETE_ALLOC_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP,
                                     ISE_DELETE_VOLUME_RESP])
        self.setup_driver()
        self.driver.delete_snapshot(SNAPSHOT1)

    def test_clone_volume(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        VOLUME1['volume_type_id'] = type_ref['id']
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_PREP_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP,
                                     ISE_CREATE_SNAPSHOT_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP])
        self.setup_driver()
        self.driver.create_cloned_volume(CLONE1, VOLUME1)

    def test_extend_volume(self, mock_req):
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_MODIFY_VOLUME_RESP])
        self.setup_driver()
        self.driver.extend_volume(VOLUME1, NEW_VOLUME_SIZE)

    def test_retype_volume(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        VOLUME1['volume_type_id'] = type_ref['id']
        # New volume type
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "5",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT2', extra_specs)
        specs = {'qos:minIOPS': '30',
                 'qos:maxIOPS': '3000',
                 'qos:burstIOPS': '10000'}
        qos = qos_specs.create(ctxt, 'fake-qos2', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])

        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_MODIFY_VOLUME_RESP])
        self.setup_driver()
        self.driver.retype(ctxt, VOLUME1, type_ref, 0, 0)

    def test_create_volume_from_snapshot(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        SNAPSHOT1['volume_type_id'] = type_ref['id']
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_SNAP1_STATUS_RESP,
                                     ISE_PREP_SNAPSHOT_RESP,
                                     ISE_GET_VOL1_STATUS_RESP,
                                     ISE_CREATE_SNAPSHOT_RESP,
                                     ISE_GET_VOL1_STATUS_RESP])
        self.setup_driver()
        self.driver.create_volume_from_snapshot(VOLUME1, SNAPSHOT1)

    def test_manage_existing(self, mock_req):
        ctxt = context.get_admin_context()
        extra_specs = {"Feature:Pool": "1",
                       "Feature:Raid": "1",
                       "Affinity:Type": "flash",
                       "Alloc:Type": "thick"}
        type_ref = volume_types.create(ctxt, 'VT1', extra_specs)
        specs = {'qos:minIOPS': '20',
                 'qos:maxIOPS': '2000',
                 'qos:burstIOPS': '5000'}
        qos = qos_specs.create(ctxt, 'fake-qos', specs)
        qos_specs.associate_qos_with_type(ctxt, qos['id'], type_ref['id'])
        VOLUME1['volume_type_id'] = type_ref['id']
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_MODIFY_VOLUME_RESP,
                                         ISE_GET_IONETWORKS_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_MODIFY_VOLUME_RESP])
        self.driver.manage_existing(VOLUME1, {'source-name': 'testvol'})

    def test_manage_existing_no_source_name(self, mock_req):
        self.setup_driver()
        if self.configuration.ise_protocol == 'iscsi':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_MODIFY_VOLUME_RESP,
                                         ISE_GET_IONETWORKS_RESP])
        elif self.configuration.ise_protocol == 'fibre_channel':
            mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                         ISE_GET_VOL1_STATUS_RESP,
                                         ISE_MODIFY_VOLUME_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.manage_existing, VOLUME1, {})

    def test_manage_existing_get_size(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP])
        exp_result = 10
        act_result = \
            self.driver.manage_existing_get_size(VOLUME1,
                                                 {'source-name': 'a'})
        self.assertEqual(exp_result, act_result)

    def test_manage_existing_get_size_no_source_name(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL1_STATUS_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.manage_existing_get_size, VOLUME1, {})

    def test_unmanage(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                    ISE_GET_VOL1_STATUS_RESP])
        self.driver.unmanage(VOLUME1)

    def test_negative_unmanage_no_volume_status_xml(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                    ISE_GET_VOL_STATUS_NO_STATUS_RESP])
        self.driver.unmanage(VOLUME1)

    def test_negative_unmanage_no_volume_xml(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                    ISE_GET_VOL_STATUS_NO_VOL_NODE_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.unmanage, VOLUME1)

    def test_negative_unmanage_non_existing_volume(self, mock_req):
        self.setup_driver()
        mock_req.side_effect = iter([ISE_GET_QUERY_RESP,
                                     ISE_GET_VOL_STATUS_404_RESP])
        self.assertRaises(exception.XIODriverException,
                          self.driver.unmanage, VOLUME1)


class XIOISEISCSIDriverTestCase(XIOISEDriverTestCase, test.TestCase):

    def setUp(self):
        super(XIOISEISCSIDriverTestCase, self).setUp()
        self.setup_test('iscsi')


class XIOISEFCDriverTestCase(XIOISEDriverTestCase, test.TestCase):

    def setUp(self):
        super(XIOISEFCDriverTestCase, self).setUp()
        self.setup_test('fibre_channel')



# Copyright 2013 IBM Corp.
# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Authors:
#   Erik Zaadi <erikz@il.ibm.com>
#   Avishay Traeger <avishay@il.ibm.com>

"""
Unified Volume driver for IBM XIV and DS8K Storage Systems.
"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from cinder import exception
from cinder.volume import driver
from cinder.volume.drivers.san import san

xiv_ds8k_opts = [
    cfg.StrOpt(
        'xiv_ds8k_proxy',
        default='xiv_ds8k_openstack.nova_proxy.XIVDS8KNovaProxy',
        help='Proxy driver that connects to the IBM Storage Array'),
    cfg.StrOpt(
        'xiv_ds8k_connection_type',
        default='iscsi',
        choices=['fibre_channel', 'iscsi'],
        help='Connection type to the IBM Storage Array'),
    cfg.StrOpt(
        'xiv_chap',
        default='disabled',
        choices=['disabled', 'enabled'],
        help='CHAP authentication mode, effective only for iscsi'
        ' (disabled|enabled)'),
    cfg.StrOpt(
        'management_ips',
        default='',
        help='List of Management IP addresses (separated by commas)'),
]

CONF = cfg.CONF
CONF.register_opts(xiv_ds8k_opts)

LOG = logging.getLogger(__name__)


class XIVDS8KDriver(san.SanDriver,
                    driver.ManageableVD,
                    driver.ExtendVD,
                    driver.SnapshotVD,
                    driver.MigrateVD,
                    driver.ConsistencyGroupVD,
                    driver.CloneableImageVD,
                    driver.TransferVD):
    """Unified IBM XIV and DS8K volume driver."""

    def __init__(self, *args, **kwargs):
        """Initialize the driver."""

        super(XIVDS8KDriver, self).__init__(*args, **kwargs)

        self.configuration.append_config_values(xiv_ds8k_opts)

        proxy = importutils.import_class(self.configuration.xiv_ds8k_proxy)

        active_backend_id = kwargs.get('active_backend_id', None)

        # NOTE: All Array specific configurations are prefixed with:
        # "xiv_ds8k_array_"
        # These additional flags should be specified in the cinder.conf
        # preferably in each backend configuration.

        self.xiv_ds8k_proxy = proxy(
            {
                "xiv_ds8k_user": self.configuration.san_login,
                "xiv_ds8k_pass": self.configuration.san_password,
                "xiv_ds8k_address": self.configuration.san_ip,
                "xiv_ds8k_vol_pool": self.configuration.san_clustername,
                "xiv_ds8k_connection_type":
                self.configuration.xiv_ds8k_connection_type,
                "xiv_chap": self.configuration.xiv_chap,
                "management_ips": self.configuration.management_ips
            },
            LOG,
            exception,
            driver=self,
            active_backend_id=active_backend_id)

    def do_setup(self, context):
        """Setup and verify IBM XIV and DS8K Storage connection."""

        self.xiv_ds8k_proxy.setup(context)

    def ensure_export(self, context, volume):
        """Ensure an export."""

        return self.xiv_ds8k_proxy.ensure_export(context, volume)

    def create_export(self, context, volume, connector):
        """Create an export."""

        return self.xiv_ds8k_proxy.create_export(context, volume)

    def create_volume(self, volume):
        """Create a volume on the IBM XIV and DS8K Storage system."""

        return self.xiv_ds8k_proxy.create_volume(volume)

    def delete_volume(self, volume):
        """Delete a volume on the IBM XIV and DS8K Storage system."""

        self.xiv_ds8k_proxy.delete_volume(volume)

    def remove_export(self, context, volume):
        """Disconnect a volume from an attached instance."""

        return self.xiv_ds8k_proxy.remove_export(context, volume)

    def initialize_connection(self, volume, connector):
        """Map the created volume."""

        return self.xiv_ds8k_proxy.initialize_connection(volume, connector)

    def terminate_connection(self, volume, connector, **kwargs):
        """Terminate a connection to a volume."""

        return self.xiv_ds8k_proxy.terminate_connection(volume, connector)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""

        return self.xiv_ds8k_proxy.create_volume_from_snapshot(
            volume,
            snapshot)

    def create_snapshot(self, snapshot):
        """Create a snapshot."""

        return self.xiv_ds8k_proxy.create_snapshot(snapshot)

    def delete_snapshot(self, snapshot):
        """Delete a snapshot."""

        return self.xiv_ds8k_proxy.delete_snapshot(snapshot)

    def get_volume_stats(self, refresh=False):
        """Get volume stats."""

        return self.xiv_ds8k_proxy.get_volume_stats(refresh)

    def create_cloned_volume(self, tgt_volume, src_volume):
        """Create Cloned Volume."""

        return self.xiv_ds8k_proxy.create_cloned_volume(tgt_volume, src_volume)

    def extend_volume(self, volume, new_size):
        """Extend Created Volume."""

        self.xiv_ds8k_proxy.extend_volume(volume, new_size)

    def migrate_volume(self, context, volume, host):
        """Migrate the volume to the specified host."""

        return self.xiv_ds8k_proxy.migrate_volume(context, volume, host)

    def manage_existing(self, volume, existing_ref):
        """Brings an existing backend storage object under Cinder management.

        existing_ref is passed straight through from the API request's
        manage_existing_ref value, and it is up to the driver how this should
        be interpreted.  It should be sufficient to identify a storage object
        that the driver should somehow associate with the newly-created cinder
        volume structure.
        In the case of XIV, the existing_ref consists of a single field named
        'existing_ref' representing the name of the volume on the storage.

        There are two ways to do this:

        1. Rename the backend storage object so that it matches the,
           volume['name'] which is how drivers traditionally map between a
           cinder volume and the associated backend storage object.

        2. Place some metadata on the volume, or somewhere in the backend, that
           allows other driver requests (e.g. delete, clone, attach, detach...)
           to locate the backend storage object when required.

        If the existing_ref doesn't make sense, or doesn't refer to an existing
        backend storage object, raise a ManageExistingInvalidReference
        exception.

        The volume may have a volume_type, and the driver can inspect that and
        compare against the properties of the referenced backend storage
        object.  If they are incompatible, raise a
        ManageExistingVolumeTypeMismatch, specifying a reason for the failure.
        """
        return self.xiv_ds8k_proxy.manage_volume(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing."""

        return self.xiv_ds8k_proxy.manage_volume_get_size(volume, existing_ref)

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management."""

        return self.xiv_ds8k_proxy.unmanage_volume(volume)

    def freeze_backend(self, context):
        """Notify the backend that it's frozen. """

        return self.xiv_ds8k_proxy.freeze_backend(context)

    def thaw_backend(self, context):
        """Notify the backend that it's unfrozen/thawed. """

        return self.xiv_ds8k_proxy.thaw_backend(context)

    def failover_host(self, context, volumes, secondary_id=None):
        """Failover a backend to a secondary replication target. """

        return self.xiv_ds8k_proxy.failover_host(
            context, volumes, secondary_id)

    def get_replication_status(self, context, volume):
        """Return replication status."""

        return self.xiv_ds8k_proxy.get_replication_status(context, volume)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type."""

        return self.xiv_ds8k_proxy.retype(ctxt, volume, new_type, diff, host)

    def create_consistencygroup(self, context, group):
        """Creates a consistency group."""

        return self.xiv_ds8k_proxy.create_consistencygroup(context, group)

    def delete_consistencygroup(self, context, group, volumes):
        """Deletes a consistency group."""

        return self.xiv_ds8k_proxy.delete_consistencygroup(
            context, group, volumes)

    def create_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Creates a consistency group snapshot."""

        return self.xiv_ds8k_proxy.create_cgsnapshot(
            context, cgsnapshot, snapshots)

    def delete_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Deletes a consistency group snapshot."""

        return self.xiv_ds8k_proxy.delete_cgsnapshot(
            context, cgsnapshot, snapshots)

    def update_consistencygroup(self, context, group,
                                add_volumes, remove_volumes):
        """Adds or removes volume(s) to/from an existing consistency group."""

        return self.xiv_ds8k_proxy.update_consistencygroup(
            context, group, add_volumes, remove_volumes)

    def create_consistencygroup_from_src(
            self, context, group, volumes, cgsnapshot, snapshots,
            source_cg=None, source_vols=None):
        """Creates a consistencygroup from source."""

        return self.xiv_ds8k_proxy.create_consistencygroup_from_src(
            context, group, volumes, cgsnapshot, snapshots,
            source_cg, source_vols)



# Copyright (c) 2015 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Column, Integer, MetaData, Table


def upgrade(migrate_engine):
    """Add num_dependent_backups column to backups."""
    meta = MetaData()
    meta.bind = migrate_engine

    backups = Table('backups', meta, autoload=True)
    num_dependent_backups = Column('num_dependent_backups', Integer, default=0)
    backups.create_column(num_dependent_backups)
    backups_list = list(backups.select().execute())
    for backup in backups_list:
        dep_bks_list = list(backups.select().where(backups.columns.parent_id ==
                                                   backup.id).execute())
        if dep_bks_list:
            backups.update().where(backups.columns.id == backup.id).values(
                num_dependent_backups=len(dep_bks_list)).execute()



# Copyright (c) 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume-related Utilities and helpers."""


import ast
import math
import re
import time
import uuid

from Crypto.Random import random
import eventlet
from eventlet import tpool
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import units
import six
from six.moves import range

from cinder.brick.local_dev import lvm as brick_lvm
from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _, _LI, _LW, _LE
from cinder import rpc
from cinder import utils
from cinder.volume import throttling


CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def null_safe_str(s):
    return str(s) if s else ''


def _usage_from_volume(context, volume_ref, **kw):
    now = timeutils.utcnow()
    launched_at = volume_ref['launched_at'] or now
    created_at = volume_ref['created_at'] or now
    usage_info = dict(
        tenant_id=volume_ref['project_id'],
        host=volume_ref['host'],
        user_id=volume_ref['user_id'],
        availability_zone=volume_ref['availability_zone'],
        volume_id=volume_ref['id'],
        volume_type=volume_ref['volume_type_id'],
        display_name=volume_ref['display_name'],
        launched_at=launched_at.isoformat(),
        created_at=created_at.isoformat(),
        status=volume_ref['status'],
        snapshot_id=volume_ref['snapshot_id'],
        size=volume_ref['size'],
        replication_status=volume_ref['replication_status'],
        replication_extended_status=volume_ref['replication_extended_status'],
        replication_driver_data=volume_ref['replication_driver_data'],
        metadata=volume_ref.get('volume_metadata'),)

    usage_info.update(kw)
    try:
        attachments = db.volume_attachment_get_used_by_volume_id(
            context, volume_ref['id'])
        usage_info['volume_attachment'] = attachments

        glance_meta = db.volume_glance_metadata_get(context, volume_ref['id'])
        if glance_meta:
            usage_info['glance_metadata'] = glance_meta
    except exception.GlanceMetadataNotFound:
        pass
    except exception.VolumeNotFound:
        LOG.debug("Can not find volume %s at notify usage", volume_ref['id'])

    return usage_info


def _usage_from_backup(backup_ref, **kw):
    num_dependent_backups = backup_ref['num_dependent_backups']
    usage_info = dict(tenant_id=backup_ref['project_id'],
                      user_id=backup_ref['user_id'],
                      availability_zone=backup_ref['availability_zone'],
                      backup_id=backup_ref['id'],
                      host=backup_ref['host'],
                      display_name=backup_ref['display_name'],
                      created_at=str(backup_ref['created_at']),
                      status=backup_ref['status'],
                      volume_id=backup_ref['volume_id'],
                      size=backup_ref['size'],
                      service_metadata=backup_ref['service_metadata'],
                      service=backup_ref['service'],
                      fail_reason=backup_ref['fail_reason'],
                      parent_id=backup_ref['parent_id'],
                      num_dependent_backups=num_dependent_backups,
                      snapshot_id=backup_ref['snapshot_id'],
                      )

    usage_info.update(kw)
    return usage_info


def notify_about_volume_usage(context, volume, event_suffix,
                              extra_usage_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_volume(context, volume, **extra_usage_info)

    rpc.get_notifier("volume", host).info(context, 'volume.%s' % event_suffix,
                                          usage_info)


def notify_about_backup_usage(context, backup, event_suffix,
                              extra_usage_info=None,
                              host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_backup(backup, **extra_usage_info)

    rpc.get_notifier("backup", host).info(context, 'backup.%s' % event_suffix,
                                          usage_info)


def _usage_from_snapshot(snapshot, **extra_usage_info):
    usage_info = {
        'tenant_id': snapshot.project_id,
        'user_id': snapshot.user_id,
        'availability_zone': snapshot.volume['availability_zone'],
        'volume_id': snapshot.volume_id,
        'volume_size': snapshot.volume_size,
        'snapshot_id': snapshot.id,
        'display_name': snapshot.display_name,
        'created_at': str(snapshot.created_at),
        'status': snapshot.status,
        'deleted': null_safe_str(snapshot.deleted),
        'metadata': null_safe_str(snapshot.metadata),
    }

    usage_info.update(extra_usage_info)
    return usage_info


def notify_about_snapshot_usage(context, snapshot, event_suffix,
                                extra_usage_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_snapshot(snapshot, **extra_usage_info)

    rpc.get_notifier('snapshot', host).info(context,
                                            'snapshot.%s' % event_suffix,
                                            usage_info)


def notify_about_replication_usage(context, volume, suffix,
                                   extra_usage_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_volume(context, volume,
                                    **extra_usage_info)

    rpc.get_notifier('replication', host).info(context,
                                               'replication.%s' % suffix,
                                               usage_info)


def notify_about_replication_error(context, volume, suffix,
                                   extra_error_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_error_info:
        extra_error_info = {}

    usage_info = _usage_from_volume(context, volume,
                                    **extra_error_info)

    rpc.get_notifier('replication', host).error(context,
                                                'replication.%s' % suffix,
                                                usage_info)


def _usage_from_consistencygroup(group_ref, **kw):
    usage_info = dict(tenant_id=group_ref.project_id,
                      user_id=group_ref.user_id,
                      availability_zone=group_ref.availability_zone,
                      consistencygroup_id=group_ref.id,
                      name=group_ref.name,
                      created_at=group_ref.created_at.isoformat(),
                      status=group_ref.status)

    usage_info.update(kw)
    return usage_info


def notify_about_consistencygroup_usage(context, group, event_suffix,
                                        extra_usage_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_consistencygroup(group,
                                              **extra_usage_info)

    rpc.get_notifier("consistencygroup", host).info(
        context,
        'consistencygroup.%s' % event_suffix,
        usage_info)


def _usage_from_cgsnapshot(cgsnapshot, **kw):
    usage_info = dict(
        tenant_id=cgsnapshot.project_id,
        user_id=cgsnapshot.user_id,
        cgsnapshot_id=cgsnapshot.id,
        name=cgsnapshot.name,
        consistencygroup_id=cgsnapshot.consistencygroup_id,
        created_at=cgsnapshot.created_at.isoformat(),
        status=cgsnapshot.status)

    usage_info.update(kw)
    return usage_info


def notify_about_cgsnapshot_usage(context, cgsnapshot, event_suffix,
                                  extra_usage_info=None, host=None):
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_cgsnapshot(cgsnapshot,
                                        **extra_usage_info)

    rpc.get_notifier("cgsnapshot", host).info(
        context,
        'cgsnapshot.%s' % event_suffix,
        usage_info)


def _calculate_count(size_in_m, blocksize):

    # Check if volume_dd_blocksize is valid
    try:
        # Rule out zero-sized/negative/float dd blocksize which
        # cannot be caught by strutils
        if blocksize.startswith(('-', '0')) or '.' in blocksize:
            raise ValueError
        bs = strutils.string_to_bytes('%sB' % blocksize)
    except ValueError:
        LOG.warning(_LW("Incorrect value error: %(blocksize)s, "
                        "it may indicate that \'volume_dd_blocksize\' "
                        "was configured incorrectly. Fall back to default."),
                    {'blocksize': blocksize})
        # Fall back to default blocksize
        CONF.clear_override('volume_dd_blocksize')
        blocksize = CONF.volume_dd_blocksize
        bs = strutils.string_to_bytes('%sB' % blocksize)

    count = math.ceil(size_in_m * units.Mi / bs)

    return blocksize, int(count)


def check_for_odirect_support(src, dest, flag='oflag=direct'):

    # Check whether O_DIRECT is supported
    try:
        # iflag=direct and if=/dev/zero combination does not work
        # error: dd: failed to open '/dev/zero': Invalid argument
        if (src == '/dev/zero' and flag == 'iflag=direct'):
            return False
        else:
            utils.execute('dd', 'count=0', 'if=%s' % src,
                          'of=%s' % dest,
                          flag, run_as_root=True)
            return True
    except processutils.ProcessExecutionError:
        return False


def _copy_volume_with_path(prefix, srcstr, deststr, size_in_m, blocksize,
                           sync=False, execute=utils.execute, ionice=None,
                           sparse=False):
    # Use O_DIRECT to avoid thrashing the system buffer cache
    extra_flags = []
    if check_for_odirect_support(srcstr, deststr, 'iflag=direct'):
        extra_flags.append('iflag=direct')

    if check_for_odirect_support(srcstr, deststr, 'oflag=direct'):
        extra_flags.append('oflag=direct')

    # If the volume is being unprovisioned then
    # request the data is persisted before returning,
    # so that it's not discarded from the cache.
    conv = []
    if sync and not extra_flags:
        conv.append('fdatasync')
    if sparse:
        conv.append('sparse')
    if conv:
        conv_options = 'conv=' + ",".join(conv)
        extra_flags.append(conv_options)

    blocksize, count = _calculate_count(size_in_m, blocksize)

    cmd = ['dd', 'if=%s' % srcstr, 'of=%s' % deststr,
           'count=%d' % count, 'bs=%s' % blocksize]
    cmd.extend(extra_flags)

    if ionice is not None:
        cmd = ['ionice', ionice] + cmd

    cmd = prefix + cmd

    # Perform the copy
    start_time = timeutils.utcnow()
    execute(*cmd, run_as_root=True)
    duration = timeutils.delta_seconds(start_time, timeutils.utcnow())

    # NOTE(jdg): use a default of 1, mostly for unit test, but in
    # some incredible event this is 0 (cirros image?) don't barf
    if duration < 1:
        duration = 1
    mbps = (size_in_m / duration)
    LOG.debug("Volume copy details: src %(src)s, dest %(dest)s, "
              "size %(sz).2f MB, duration %(duration).2f sec",
              {"src": srcstr,
               "dest": deststr,
               "sz": size_in_m,
               "duration": duration})
    LOG.info(_LI("Volume copy %(size_in_m).2f MB at %(mbps).2f MB/s"),
             {'size_in_m': size_in_m, 'mbps': mbps})


def _open_volume_with_path(path, mode):
    try:
        with utils.temporary_chown(path):
            handle = open(path, mode)
            return handle
    except Exception:
        LOG.error(_LE("Failed to open volume from %(path)s."), {'path': path})


def _transfer_data(src, dest, length, chunk_size):
    """Transfer data between files (Python IO objects)."""

    chunks = int(math.ceil(length / chunk_size))
    remaining_length = length

    LOG.debug("%(chunks)s chunks of %(bytes)s bytes to be transferred.",
              {'chunks': chunks, 'bytes': chunk_size})

    for chunk in range(0, chunks):
        before = time.time()
        data = tpool.execute(src.read, min(chunk_size, remaining_length))

        # If we have reached end of source, discard any extraneous bytes from
        # destination volume if trim is enabled and stop writing.
        if data == b'':
            break

        tpool.execute(dest.write, data)
        remaining_length -= len(data)
        delta = (time.time() - before)
        rate = (chunk_size / delta) / units.Ki
        LOG.debug("Transferred chunk %(chunk)s of %(chunks)s (%(rate)dK/s).",
                  {'chunk': chunk + 1, 'chunks': chunks, 'rate': rate})

        # yield to any other pending operations
        eventlet.sleep(0)

    tpool.execute(dest.flush)


def _copy_volume_with_file(src, dest, size_in_m):
    src_handle = src
    if isinstance(src, six.string_types):
        src_handle = _open_volume_with_path(src, 'rb')

    dest_handle = dest
    if isinstance(dest, six.string_types):
        dest_handle = _open_volume_with_path(dest, 'wb')

    if not src_handle:
        raise exception.DeviceUnavailable(
            _("Failed to copy volume, source device unavailable."))

    if not dest_handle:
        raise exception.DeviceUnavailable(
            _("Failed to copy volume, destination device unavailable."))

    start_time = timeutils.utcnow()

    _transfer_data(src_handle, dest_handle, size_in_m * units.Mi, units.Mi * 4)

    duration = max(1, timeutils.delta_seconds(start_time, timeutils.utcnow()))

    if isinstance(src, six.string_types):
        src_handle.close()
    if isinstance(dest, six.string_types):
        dest_handle.close()

    mbps = (size_in_m / duration)
    LOG.info(_LI("Volume copy completed (%(size_in_m).2f MB at "
                 "%(mbps).2f MB/s)."),
             {'size_in_m': size_in_m, 'mbps': mbps})


def copy_volume(src, dest, size_in_m, blocksize, sync=False,
                execute=utils.execute, ionice=None, throttle=None,
                sparse=False):
    """Copy data from the source volume to the destination volume.

    The parameters 'src' and 'dest' are both typically of type str, which
    represents the path to each volume on the filesystem.  Connectors can
    optionally return a volume handle of type RawIOBase for volumes that are
    not available on the local filesystem for open/close operations.

    If either 'src' or 'dest' are not of type str, then they are assumed to be
    of type RawIOBase or any derivative that supports file operations such as
    read and write.  In this case, the handles are treated as file handles
    instead of file paths and, at present moment, throttling is unavailable.
    """

    if (isinstance(src, six.string_types) and
            isinstance(dest, six.string_types)):
        if not throttle:
            throttle = throttling.Throttle.get_default()
        with throttle.subcommand(src, dest) as throttle_cmd:
            _copy_volume_with_path(throttle_cmd['prefix'], src, dest,
                                   size_in_m, blocksize, sync=sync,
                                   execute=execute, ionice=ionice,
                                   sparse=sparse)
    else:
        _copy_volume_with_file(src, dest, size_in_m)


def clear_volume(volume_size, volume_path, volume_clear=None,
                 volume_clear_size=None, volume_clear_ionice=None,
                 throttle=None):
    """Unprovision old volumes to prevent data leaking between users."""
    if volume_clear is None:
        volume_clear = CONF.volume_clear

    if volume_clear_size is None:
        volume_clear_size = CONF.volume_clear_size

    if volume_clear_size == 0:
        volume_clear_size = volume_size

    if volume_clear_ionice is None:
        volume_clear_ionice = CONF.volume_clear_ionice

    LOG.info(_LI("Performing secure delete on volume: %s"), volume_path)

    # We pass sparse=False explicitly here so that zero blocks are not
    # skipped in order to clear the volume.
    if volume_clear == 'zero':
        return copy_volume('/dev/zero', volume_path, volume_clear_size,
                           CONF.volume_dd_blocksize,
                           sync=True, execute=utils.execute,
                           ionice=volume_clear_ionice,
                           throttle=throttle, sparse=False)
    elif volume_clear == 'shred':
        clear_cmd = ['shred', '-n3']
        if volume_clear_size:
            clear_cmd.append('-s%dMiB' % volume_clear_size)
    else:
        raise exception.InvalidConfigurationValue(
            option='volume_clear',
            value=volume_clear)

    clear_cmd.append(volume_path)
    start_time = timeutils.utcnow()
    utils.execute(*clear_cmd, run_as_root=True)
    duration = timeutils.delta_seconds(start_time, timeutils.utcnow())

    # NOTE(jdg): use a default of 1, mostly for unit test, but in
    # some incredible event this is 0 (cirros image?) don't barf
    if duration < 1:
        duration = 1
    LOG.info(_LI('Elapsed time for clear volume: %.2f sec'), duration)


def supports_thin_provisioning():
    return brick_lvm.LVM.supports_thin_provisioning(
        utils.get_root_helper())


def get_all_physical_volumes(vg_name=None):
    return brick_lvm.LVM.get_all_physical_volumes(
        utils.get_root_helper(),
        vg_name)


def get_all_volume_groups(vg_name=None):
    return brick_lvm.LVM.get_all_volume_groups(
        utils.get_root_helper(),
        vg_name)

# Default symbols to use for passwords. Avoids visually confusing characters.
# ~6 bits per symbol
DEFAULT_PASSWORD_SYMBOLS = ('23456789',  # Removed: 0,1
                            'ABCDEFGHJKLMNPQRSTUVWXYZ',   # Removed: I, O
                            'abcdefghijkmnopqrstuvwxyz')  # Removed: l


def generate_password(length=16, symbolgroups=DEFAULT_PASSWORD_SYMBOLS):
    """Generate a random password from the supplied symbol groups.

    At least one symbol from each group will be included. Unpredictable
    results if length is less than the number of symbol groups.

    Believed to be reasonably secure (with a reasonable password length!)

    """
    # NOTE(jerdfelt): Some password policies require at least one character
    # from each group of symbols, so start off with one random character
    # from each symbol group
    password = [random.choice(s) for s in symbolgroups]
    # If length < len(symbolgroups), the leading characters will only
    # be from the first length groups. Try our best to not be predictable
    # by shuffling and then truncating.
    random.shuffle(password)
    password = password[:length]
    length -= len(password)

    # then fill with random characters from all symbol groups
    symbols = ''.join(symbolgroups)
    password.extend([random.choice(symbols) for _i in range(length)])

    # finally shuffle to ensure first x characters aren't from a
    # predictable group
    random.shuffle(password)

    return ''.join(password)


def generate_username(length=20, symbolgroups=DEFAULT_PASSWORD_SYMBOLS):
    # Use the same implementation as the password generation.
    return generate_password(length, symbolgroups)


DEFAULT_POOL_NAME = '_pool0'


def extract_host(host, level='backend', default_pool_name=False):
    """Extract Host, Backend or Pool information from host string.

    :param host: String for host, which could include host@backend#pool info
    :param level: Indicate which level of information should be extracted
                  from host string. Level can be 'host', 'backend' or 'pool',
                  default value is 'backend'
    :param default_pool_name: this flag specify what to do if level == 'pool'
                              and there is no 'pool' info encoded in host
                              string.  default_pool_name=True will return
                              DEFAULT_POOL_NAME, otherwise we return None.
                              Default value of this parameter is False.
    :return: expected level of information

    For example:
        host = 'HostA@BackendB#PoolC'
        ret = extract_host(host, 'host')
        # ret is 'HostA'
        ret = extract_host(host, 'backend')
        # ret is 'HostA@BackendB'
        ret = extract_host(host, 'pool')
        # ret is 'PoolC'

        host = 'HostX@BackendY'
        ret = extract_host(host, 'pool')
        # ret is None
        ret = extract_host(host, 'pool', True)
        # ret is '_pool0'
    """
    if level == 'host':
        # make sure pool is not included
        hst = host.split('#')[0]
        return hst.split('@')[0]
    elif level == 'backend':
        return host.split('#')[0]
    elif level == 'pool':
        lst = host.split('#')
        if len(lst) == 2:
            return lst[1]
        elif default_pool_name is True:
            return DEFAULT_POOL_NAME
        else:
            return None


def get_volume_rpc_host(host):
    if CONF.rpc_backend and CONF.rpc_backend == "zmq":
        # ZeroMQ RPC driver requires only the hostname.
        # So, return just that.
        return extract_host(host, 'host')
    return extract_host(host)


def append_host(host, pool):
    """Encode pool into host info."""
    if not host or not pool:
        return host

    new_host = "#".join([host, pool])
    return new_host


def matching_backend_name(src_volume_type, volume_type):
    if src_volume_type.get('volume_backend_name') and \
            volume_type.get('volume_backend_name'):
        return src_volume_type.get('volume_backend_name') == \
            volume_type.get('volume_backend_name')
    else:
        return False


def hosts_are_equivalent(host_1, host_2):
    return extract_host(host_1) == extract_host(host_2)


def read_proc_mounts():
    """Read the /proc/mounts file.

    It's a dummy function but it eases the writing of unit tests as mocking
    __builtin__open() for a specific file only is not trivial.
    """
    with open('/proc/mounts') as mounts:
        return mounts.readlines()


def _extract_id(vol_name):
    regex = re.compile(
        CONF.volume_name_template.replace('%s', '(?P<uuid>.+)'))
    match = regex.match(vol_name)
    return match.group('uuid') if match else None


def check_already_managed_volume(db, vol_name):
    """Check cinder db for already managed volume.

    :param db: database api parameter
    :param vol_name: volume name parameter
    :returns: bool -- return True, if db entry with specified
                      volume name exist, otherwise return False
    """
    vol_id = _extract_id(vol_name)
    try:
        if vol_id and uuid.UUID(vol_id, version=4):
            db.volume_get(context.get_admin_context(), vol_id)
            return True
    except (exception.VolumeNotFound, ValueError):
        return False
    return False


def convert_config_string_to_dict(config_string):
    """Convert config file replication string to a dict.

    The only supported form is as follows:
    "{'key-1'='val-1' 'key-2'='val-2'...}"

    :param config_string: Properly formatted string to convert to dict.
    :response: dict of string values
    """

    resultant_dict = {}

    try:
        st = config_string.replace("=", ":")
        st = st.replace(" ", ", ")
        resultant_dict = ast.literal_eval(st)
    except Exception:
        LOG.warning(_LW("Error encountered translating config_string: "
                        "%(config_string)s to dict"),
                    {'config_string': config_string})

    return resultant_dict


def process_reserve_over_quota(context, overs, usages, quotas, size):
    def _consumed(name):
        return (usages[name]['reserved'] + usages[name]['in_use'])

    for over in overs:
        if 'gigabytes' in over:
            msg = _LW("Quota exceeded for %(s_pid)s, tried to create "
                      "%(s_size)sG snapshot (%(d_consumed)dG of "
                      "%(d_quota)dG already consumed).")
            LOG.warning(msg, {'s_pid': context.project_id,
                              's_size': size,
                              'd_consumed': _consumed(over),
                              'd_quota': quotas[over]})
            raise exception.VolumeSizeExceedsAvailableQuota(
                requested=size,
                consumed=_consumed('gigabytes'),
                quota=quotas['gigabytes'])
        elif 'snapshots' in over:
            msg = _LW("Quota exceeded for %(s_pid)s, tried to create "
                      "snapshot (%(d_consumed)d snapshots "
                      "already consumed).")
            LOG.warning(msg, {'s_pid': context.project_id,
                              'd_consumed': _consumed(over)})
            raise exception.SnapshotLimitExceeded(allowed=quotas[over])



# Copyright 2015 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from cinder import exception
from cinder.i18n import _, _LE
from cinder.volume import driver
from cinder.volume.drivers.nexenta.nexentaedge import jsonrpc
from cinder.volume.drivers.nexenta import options


LOG = logging.getLogger(__name__)


class NexentaEdgeISCSIDriver(driver.ISCSIDriver):
    """Executes volume driver commands on NexentaEdge cluster.

    Version history:
        1.0.0 - Initial driver version.
        1.0.1 - Moved opts to options.py.
    """

    VERSION = '1.0.1'

    def __init__(self, *args, **kwargs):
        super(NexentaEdgeISCSIDriver, self).__init__(*args, **kwargs)
        if self.configuration:
            self.configuration.append_config_values(
                options.NEXENTA_CONNECTION_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_ISCSI_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_DATASET_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_EDGE_OPTS)
        self.restapi_protocol = self.configuration.nexenta_rest_protocol
        self.restapi_host = self.configuration.nexenta_rest_address
        self.restapi_port = self.configuration.nexenta_rest_port
        self.restapi_user = self.configuration.nexenta_rest_user
        self.restapi_password = self.configuration.nexenta_rest_password
        self.iscsi_service = self.configuration.nexenta_iscsi_service
        self.bucket_path = self.configuration.nexenta_lun_container
        self.blocksize = self.configuration.nexenta_blocksize
        self.chunksize = self.configuration.nexenta_chunksize
        self.cluster, self.tenant, self.bucket = self.bucket_path.split('/')
        self.bucket_url = ('clusters/' + self.cluster + '/tenants/' +
                           self.tenant + '/buckets/' + self.bucket)
        self.iscsi_target_port = (self.configuration.
                                  nexenta_iscsi_target_portal_port)
        self.target_vip = None

    @property
    def backend_name(self):
        backend_name = None
        if self.configuration:
            backend_name = self.configuration.safe_get('volume_backend_name')
        if not backend_name:
            backend_name = self.__class__.__name__
        return backend_name

    def do_setup(self, context):
        if self.restapi_protocol == 'auto':
            protocol, auto = 'http', True
        else:
            protocol, auto = self.restapi_protocol, False

        try:
            self.restapi = jsonrpc.NexentaEdgeJSONProxy(
                protocol, self.restapi_host, self.restapi_port, '/',
                self.restapi_user, self.restapi_password, auto=auto)

            rsp = self.restapi.get('service/'
                                   + self.iscsi_service + '/iscsi/status')
            data_keys = rsp['data'][list(rsp['data'].keys())[0]]
            self.target_name = data_keys.split('\n', 1)[0].split(' ')[2]

            rsp = self.restapi.get('service/' + self.iscsi_service)
            if 'X-VIPS' in rsp['data']:
                vips = json.loads(rsp['data']['X-VIPS'])
                if len(vips[0]) == 1:
                    self.target_vip = vips[0][0]['ip'].split('/', 1)[0]
                else:
                    self.target_vip = vips[0][1]['ip'].split('/', 1)[0]
            else:
                self.target_vip = self.configuration.safe_get(
                    'nexenta_client_address')
                if not self.target_vip:
                    LOG.error(_LE('No VIP configured for service %s'),
                              self.iscsi_service)
                    raise exception.VolumeBackendAPIException(
                        _('No service VIP configured and '
                          'no nexenta_client_address'))
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error verifying iSCSI service %(serv)s on '
                              'host %(hst)s'), {'serv': self.iscsi_service,
                              'hst': self.restapi_host})

    def check_for_setup_error(self):
        try:
            self.restapi.get(self.bucket_url + '/objects/')
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error verifying LUN container %(bkt)s'),
                              {'bkt': self.bucket_path})

    def _get_lun_number(self, volname):
        try:
            rsp = self.restapi.put(
                'service/' + self.iscsi_service + '/iscsi/number',
                {
                    'objectPath': self.bucket_path + '/' + volname
                })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error retrieving LUN %(vol)s number'),
                              {'vol': volname})

        return rsp['data']

    def _get_target_address(self, volname):
        return self.target_vip

    def _get_provider_location(self, volume):
        return '%(host)s:%(port)s,1 %(name)s %(number)s' % {
            'host': self._get_target_address(volume['name']),
            'port': self.iscsi_target_port,
            'name': self.target_name,
            'number': self._get_lun_number(volume['name'])
        }

    def create_volume(self, volume):
        try:
            self.restapi.post('service/' + self.iscsi_service + '/iscsi', {
                'objectPath': self.bucket_path + '/' + volume['name'],
                'volSizeMB': int(volume['size']) * units.Ki,
                'blockSize': self.blocksize,
                'chunkSize': self.chunksize
            })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error creating volume'))

    def delete_volume(self, volume):
        try:
            self.restapi.delete('service/' + self.iscsi_service +
                                '/iscsi', {'objectPath': self.bucket_path +
                                           '/' + volume['name']})
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error deleting volume'))

    def extend_volume(self, volume, new_size):
        try:
            self.restapi.put('service/' + self.iscsi_service + '/iscsi/resize',
                             {'objectPath': self.bucket_path +
                              '/' + volume['name'],
                              'newSizeMB': new_size * units.Ki})
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error extending volume'))

    def create_volume_from_snapshot(self, volume, snapshot):
        try:
            self.restapi.put(
                'service/' + self.iscsi_service + '/iscsi/snapshot/clone',
                {
                    'objectPath': self.bucket_path + '/' +
                    snapshot['volume_name'],
                    'clonePath': self.bucket_path + '/' + volume['name'],
                    'snapName': snapshot['name']
                })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error cloning volume'))

    def create_snapshot(self, snapshot):
        try:
            self.restapi.post(
                'service/' + self.iscsi_service + '/iscsi/snapshot',
                {
                    'objectPath': self.bucket_path + '/' +
                    snapshot['volume_name'],
                    'snapName': snapshot['name']
                })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error creating snapshot'))

    def delete_snapshot(self, snapshot):
        try:
            self.restapi.delete(
                'service/' + self.iscsi_service + '/iscsi/snapshot',
                {
                    'objectPath': self.bucket_path + '/' +
                    snapshot['volume_name'],
                    'snapName': snapshot['name']
                })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error deleting snapshot'))

    def create_cloned_volume(self, volume, src_vref):
        vol_url = (self.bucket_url + '/objects/' +
                   src_vref['name'] + '/clone')
        clone_body = {
            'tenant_name': self.tenant,
            'bucket_name': self.bucket,
            'object_name': volume['name']
        }
        try:
            self.restapi.post(vol_url, clone_body)
            self.restapi.post('service/' + self.iscsi_service + '/iscsi', {
                'objectPath': self.bucket_path + '/' + volume['name'],
                'volSizeMB': int(src_vref['size']) * units.Ki,
                'blockSize': self.blocksize,
                'chunkSize': self.chunksize
            })
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE('Error creating cloned volume'))

    def create_export(self, context, volume, connector=None):
        return {'provider_location': self._get_provider_location(volume)}

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    def local_path(self, volume):
        raise NotImplementedError

    def get_volume_stats(self, refresh=False):
        location_info = '%(driver)s:%(host)s:%(bucket)s' % {
            'driver': self.__class__.__name__,
            'host': self._get_target_address(None),
            'bucket': self.bucket_path
        }
        return {
            'vendor_name': 'Nexenta',
            'driver_version': self.VERSION,
            'storage_protocol': 'iSCSI',
            'reserved_percentage': 0,
            'total_capacity_gb': 'unknown',
            'free_capacity_gb': 'unknown',
            'QoS_support': False,
            'volume_backend_name': self.backend_name,
            'location_info': location_info,
            'iscsi_target_portal_port': self.iscsi_target_port,
            'restapi_url': self.restapi.url
        }
